#include <windows.h>
#include <stdio.h>

#define EI_NIDENT 16

typedef struct{
    ULONG   p_type;
    ULONG   p_offset;
    ULONG   p_vaddr;
    ULONG   p_paddr;
    ULONG   p_filesz;
    ULONG   p_memsz;
    ULONG   p_flags;
    ULONG   p_align;
}Elf32_Phdr;

typedef struct{
    UCHAR   e_ident[EI_NIDENT];
    USHORT  e_type;
    USHORT  e_machine;
    ULONG   e_version;
    ULONG   e_entry;
    ULONG   e_phoff;
    ULONG   e_shoff;
    ULONG   e_flags;
    USHORT  e_ehsize;
    USHORT  e_phentsize;
    USHORT  e_phnum;
    USHORT  e_shentsize;
    USHORT  e_shnum;
    USHORT  e_shstrndx;
    Elf32_Phdr p[10];
}Elf32_Ehdr;

typedef struct{
    ULONG   sh_name;
    ULONG   sh_type;
    ULONG   sh_flags;
    ULONG   sh_addr;
    ULONG   sh_offset;
    ULONG   sh_size;
    ULONG   sh_link;
    ULONG   sh_info;
    ULONG   sh_addralign;
    ULONG   sh_entsize;
}Elf32_Shdr;

typedef struct{
    ULONG   st_name;
    ULONG   st_value;
    ULONG   st_size;
    UCHAR   st_info;
    UCHAR   st_other;
    USHORT  st_shndx;
}Elf32_Sym;


enum {
    SHT_NULL     = 0,
    SHT_PROGBITS = 1,
    SHT_SYMTAB   = 2,
    SHT_STRTAB   = 3,
    SHT_RELA     = 4,
    SHT_HASH     = 5,
    SHT_DYNAMIC  = 6,
    SHT_NOTE     = 7,
    SHT_NOBITS   = 8,
    SHT_REL      = 9,
    SHT_SHLIB    = 10,
    SHT_DYNSYM   = 11,
    SHT_LOPROC   = 0x70000000,
    SHT_HIPROC   = 0x7fffffff,
    SHT_LOUSER   = 0x80000000,
    SHT_HIUSER   = 0xffffffff,
};







#define to_printable(c) ((c>0x20) && (c<0x80) ? c : ' ')
#define ntohs(d) (((d>>8)&0x00ff) | ((d<<8)&0xff00))
#define ntohl(d) (((d>>24)&0x0000ff) | ((d>>8)&0x0000ff00) | ((d<<8)&0x00ff0000) | ((d<<24)&0xff000000))

#define MAX_FBUF 0x400000       //4mbyte
UCHAR fbuf[MAX_FBUF];

UCHAR hbuf[0x400];
UCHAR sbuf[0x400];
UCHAR strtab[0x40000];
UCHAR tbuf[0x400];

void dump(UCHAR *buf, int len)
{
    int i,j;

    for (i=0; i<len; i+=0x10) {
        printf("\n%04x : ", i);
        for (j=0; j<0x10; j++) {
            printf("%02x ", buf[i+j]);
        }
        printf("  ");
        for (j=0; j<0x10; j++) {
            printf("%c", to_printable(buf[i+j]));
        }
    }
}

main()
{
    FILE *Ip;
    int fileno;
    int flen;
    Elf32_Ehdr *elf;
    Elf32_Phdr *p;
    Elf32_Shdr *s;
    int i;
    int ret;
    FILE *Op;

    Ip=fopen("vmlinux", "rb");
    if (!Ip) {
        printf("\nconnot open file");
        return -1;
    }
    fileno = _fileno(Ip);
    flen = _filelength(fileno);

    fread(hbuf, 0x400, 1, Ip);
    memset(&fbuf[0], 0, flen);

    elf = (Elf32_Ehdr *)&hbuf[0];
    printf("\ne_ident-magic   = %08x %c%c%c%c", ntohl(*(ULONG *)(&elf->e_ident[0])), to_printable(elf->e_ident[0]), to_printable(elf->e_ident[1]), to_printable(elf->e_ident[2]), to_printable(elf->e_ident[3]));
    printf("\ne_ident-class   = %02x", elf->e_ident[4]);
    printf("\ne_ident-data    = %02x", elf->e_ident[5]);
    printf("\ne_ident-version = %02x", elf->e_ident[6]);
    printf("\ne_type          = %04x", ntohs(elf->e_type));
    printf("\ne_machine       = %04x", ntohs(elf->e_machine));
    printf("\ne_version       = %08x", ntohl(elf->e_version));
    printf("\ne_entry         = %08x", ntohl(elf->e_entry));
    printf("\ne_phoff         = %08x", ntohl(elf->e_phoff));
    printf("\ne_shoff         = %08x", ntohl(elf->e_shoff));
    printf("\ne_flags         = %08x", ntohl(elf->e_flags));
    printf("\ne_ehsize        = %04x", ntohs(elf->e_ehsize));
    printf("\ne_phentsize     = %04x", ntohs(elf->e_phentsize));
    printf("\ne_phnum         = %04x", ntohs(elf->e_phnum));
    printf("\ne_shentsize     = %04x", ntohs(elf->e_shentsize));
    printf("\ne_shnum         = %04x", ntohs(elf->e_shnum));
    printf("\ne_shstrndx      = %04x", ntohs(elf->e_shstrndx));

    if ((elf->e_ident[0] != 0x7f) && (elf->e_ident[1] != 'E') && (elf->e_ident[2] != 'L') && (elf->e_ident[3] != 'F')) {
        printf("\nnot an elf file");
        return -1;
    }

    if (elf->e_ident[4] != 0x01) {
        printf("\nnot a 32 bit class");
        return -1;
    }

    if (elf->e_ident[5] != 0x02) {
        printf("\nnot in big endian format");
        return -1;
    }

    if (elf->e_ident[6] != 0x01) {
        printf("\nnot a valid version");
        return -1;
    }

    if (ntohs(elf->e_machine) != 8) {
        printf("\nnot a mips file");
        return -1;
    }

    if (ntohl(elf->e_version) != 1) {
        printf("\nnot a current version");
        return -1;
    }


    for (i=0; i<ntohs(elf->e_phnum); i++) {
        p = (Elf32_Phdr *)&hbuf[ntohl(elf->e_phoff)+i*ntohs(elf->e_phentsize)];
        printf("\n");
        printf("\nprogram header %d, %x", i, ntohl(elf->e_phoff)+i*ntohs(elf->e_phentsize));
        printf("\np_type       = %08x", ntohl(elf->p[i].p_type));
        printf("\np_offset     = %08x", ntohl(elf->p[i].p_offset));
        printf("\np_vaddr      = %08x", ntohl(elf->p[i].p_vaddr));
        printf("\np_paddr      = %08x", ntohl(elf->p[i].p_paddr));
        printf("\np_filesz     = %08x", ntohl(elf->p[i].p_filesz));
        printf("\np_memsz      = %08x", ntohl(elf->p[i].p_memsz));
        printf("\np_flags      = %08x", ntohl(elf->p[i].p_flags));
        printf("\np_align      = %08x", ntohl(elf->p[i].p_align));
    }

    ret = fseek(Ip, ntohl(elf->e_shoff), SEEK_SET);
    ret = fread(&sbuf[0], 0x400, 1, Ip);

    //get the string table
    s = (Elf32_Shdr *)&sbuf[ntohs(elf->e_shstrndx)*ntohs(elf->e_shentsize)];
    if (ntohl(s->sh_type) == SHT_STRTAB) {
        ret = fseek(Ip, ntohl(s->sh_offset), SEEK_SET);
        ret = fread(&strtab[0], ntohl(s->sh_size), 1, Ip);
//        dump(&strtab[0], ntohl(s->sh_size));
    }

    for (i=0; i<ntohs(elf->e_shnum); i++) {
        s = (Elf32_Shdr *)&sbuf[i*ntohs(elf->e_shentsize)];
        printf("\n");
        printf("\nsection header %d, %x", i, i*ntohs(elf->e_shentsize));
        printf("\nsh_name      = %08x, %s", ntohl(s->sh_name), &strtab[ntohl(s->sh_name)]);
        printf("\nsh_type      = %08x", ntohl(s->sh_type));
        printf("\nsh_flags     = %08x", ntohl(s->sh_flags));
        printf("\nsh_addr      = %08x", ntohl(s->sh_addr));
        printf("\nsh_offset    = %08x", ntohl(s->sh_offset));
        printf("\nsh_size      = %08x", ntohl(s->sh_size));
        printf("\nsh_link      = %08x", ntohl(s->sh_link));
        printf("\nsh_info      = %08x", ntohl(s->sh_info));
        printf("\nsh_addralign = %08x", ntohl(s->sh_addralign));
        printf("\nsh_entsize   = %08x", ntohl(s->sh_entsize));

        fseek(Ip, ntohl(s->sh_offset), SEEK_SET);
        fread(&tbuf[0], 0x40, 1, Ip);
        dump(&tbuf[0], 0x40);

        if ((ntohl(s->sh_type) == SHT_PROGBITS) && (ntohl(s->sh_addr) >= ntohl(elf->p[0].p_paddr)) && (ntohl(s->sh_size) > 0)) {
            printf("\nload file: %x+%x bytes to mem: %x+%x", ntohl(s->sh_offset), ntohl(s->sh_size), ntohl(s->sh_addr), ntohl(s->sh_size));

            printf("\nfseek(%x, %x, %x)", Ip, ntohl(s->sh_offset), SEEK_SET);
            ret = fseek(Ip, ntohl(s->sh_offset), SEEK_SET);
            printf(", ret=%x", ret);

            printf("\nfread(&fbuf[%x], %x, %x, %x)", ntohl(s->sh_addr)-ntohl(elf->p[0].p_paddr), ntohl(s->sh_size), 1, Ip);
            ret = fread(&fbuf[ntohl(s->sh_addr)-ntohl(elf->p[0].p_paddr)], ntohl(s->sh_size), 1, Ip);
            printf(", ret=%x", ret);
            if (ret == 0) {
                printf("fread failed");
                fclose(Ip);
                return -1;
            }
//            dump(&fbuf[ntohl(s->sh_addr)-ntohl(elf->p[0].p_paddr)], 0x40);

            flen = ntohl(s->sh_addr)-ntohl(elf->p[0].p_paddr)+ntohl(s->sh_size);
        }
    }

    fclose(Ip);


    printf("\nflen=%x",flen);
//    dump(&fbuf[0], flen);

    Op = fopen("vmlinux.bin", "wb");
    if (!Op) {
        printf("\ncannot open file");
        return -1;
    }
    fwrite(&fbuf[0], flen, 1, Op);
    fclose(Op);

    return 0;
}
